from __future__ import annotations

from .configuration import BaseConfig
from .vcxproj import VcxProj

from dataclasses import dataclass, field
from pathlib import Path
import re
from typing import Optional
import uuid


@dataclass
class ProjectData:

    uuid: str
    name: str
    path: Path

    configurations: set[BaseConfig] = field(default_factory=set)
    subdir: Optional[str] = None
    build_solution_target: bool = False

    def __post_init__(self):
        if self.subdir:
            self.subdir = self.subdir.strip("/")

    def __lt__(self, other: ProjectData) -> bool:
        return self.name < other.name


class SolutionFile:

    VCXPROJ_UUID = "{8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942}"
    SUBDIR_UUID = "{2150E333-8FDC-42A3-9474-1A3956D46DE8}"

    UUID_RE = r'\{[A-F0-9]{8}-[A-F0-9]{4}-[A-F0-9]{4}-[A-F0-9]{4}-[A-F0-9]{12}\}'

    def __init__(self, solution_path: Path):
        self.path = solution_path.with_suffix('.sln')

        self.projects: dict[str, ProjectData] = {}

        self._subdirs: dict[str, str] = {}  # map path to uuid
        self.uuid = f"{{{str(uuid.uuid4())}}}".upper()

    def add_project(self, project: VcxProj, subdir: Optional[str] = None, build_solution_target: bool = False) -> None:
        project_path = project.filepath
        if project_path.is_relative_to(self.path.parent):
            project_path = project_path.relative_to(self.path.parent)

        project_data = ProjectData(
            project.uuid, project.name, project_path, subdir=subdir, build_solution_target=build_solution_target
        )

        if subdir:
            self._add_subdir(subdir.strip('/'))

        project_data.configurations.update(BaseConfig(c.name, c.arch) for c in project.configs.values())
        self.projects[project.uuid] = project_data

    def _add_subdir(self, subdir: str) -> None:
        parts = subdir.split('/')
        while parts:
            subpath = '/'.join(parts)
            if subpath not in self._subdirs:
                self._subdirs[subpath] = f"{{{str(uuid.uuid5(uuid.NAMESPACE_URL, subpath))}}}".upper()
            parts.pop(-1)

    def load_configs(self) -> None:
        if not self.path.exists():
            return

        contents = self.path.read_text(encoding='utf-8')
        m = re.search(f'SolutionGuid = ({self.UUID_RE})', contents)
        self.uuid = m[1]

        for m in re.finditer(rf'Project\("{self.VCXPROJ_UUID}"\) = "(.*)", "(.*)", "({self.UUID_RE})"', contents):
            project_path = Path(m[2])
            if (self.path.parent / project_path).with_suffix(VcxProj.EXT).exists():
                project_data = ProjectData(m[3], m[1], project_path)
                self.projects[project_data.uuid] = project_data

        sln_subdirs = {}  # map subdir uuid to subdir name
        for m in re.finditer(rf'Project\("{self.SUBDIR_UUID}"\) = "(.*)", "(.*)", "({self.UUID_RE})"', contents):
            sln_subdirs[m[3]] = m[1]

        # rebuild solution hieararchy...
        sln_subdirs_parents = {}
        prj_subdirs = {}  # map project uuid to subdir uuid
        for m in re.finditer(f'({self.UUID_RE}) = ({self.UUID_RE})', contents):
            if m[1] in sln_subdirs:
                sln_subdirs_parents[m[1]] = m[2]
            elif m[1] in self.projects:
                prj_subdirs[m[1]] = m[2]

        while sln_subdirs:
            top_level = {u for u in sln_subdirs if u not in sln_subdirs_parents}

            for u, p in list(sln_subdirs_parents.items()):
                if p in top_level:
                    sln_subdirs[u] = sln_subdirs[p] + '/' + sln_subdirs[u]
                    del sln_subdirs_parents[u]

            for t in top_level:
                sd = sln_subdirs.pop(t)
                self._subdirs[sd] = t
                
                for p, u in list(prj_subdirs.items()):
                    if u == t:
                        self.projects[p].subdir = sd
                        del prj_subdirs[p]
        assert not sln_subdirs_parents

        for m in re.finditer(r'(\{[0-9A-F-]+\}).([\w-]+\|[\w]+)\.ActiveCfg', contents):
            self.projects[m[1]].configurations.add(BaseConfig.from_str(m[2]))
        for m in re.finditer(r'(\{[0-9A-F-]+\}).([\w-]+\|[\w]+)\.Build\.0', contents):
            self.projects[m[1]].build_solution_target = True

    def write(self) -> None:
        # Reference: https://learn.microsoft.com/en-us/visualstudio/extensibility/internals/solution-dot-sln-file?view=vs-2022  # noqa
        contents = [  # FIXME: versions here are totally arbitrary...
            'Microsoft Visual Studio Solution File, Format Version 12.00',
            '# Visual Studio Version 16',
            'VisualStudioVersion = 16.0.30204.135',
            'MinimumVisualStudioVersion = 10.0.40219.1',
        ]

        configurations = set()
        for project in sorted(self.projects.values()):
            contents.append(f'Project("{self.VCXPROJ_UUID}") = "{project.name}", "{project.path}", "{project.uuid}"')
            contents.append('EndProject')
            configurations.update(project.configurations)
        for subdir in sorted(self._subdirs):
            subdir_uuid = self._subdirs[subdir]
            parts = subdir.split('/')
            contents.append(f'Project("{self.SUBDIR_UUID}") = "{parts[-1]}", "{parts[-1]}", "{subdir_uuid}"')
            contents.append('EndProject')

        contents.append('Global')

        contents.append('	GlobalSection(SolutionConfigurationPlatforms) = preSolution')
        for config in sorted(configurations):
            contents.append(f'		{config} = {config}')
        contents.append('	EndGlobalSection')

        contents.append('	GlobalSection(ProjectConfigurationPlatforms) = postSolution')
        for project in sorted(self.projects.values()):
            for config in sorted(project.configurations):
                contents.append(f'		{project.uuid}.{config}.ActiveCfg = {config}')
                if project.build_solution_target:
                    contents.append(f'		{project.uuid}.{config}.Build.0 = {config}')
        contents.append('	EndGlobalSection')

        contents.append('	GlobalSection(SolutionProperties) = preSolution')
        contents.append('		HideSolutionNode = FALSE')
        contents.append('	EndGlobalSection')

        if self._subdirs:
            contents.append('	GlobalSection(NestedProjects) = preSolution')

            for project in sorted(self.projects.values()):
                if project.subdir:
                    subdir_uuid = self._subdirs[project.subdir]
                    contents.append(f'		{project.uuid} = {subdir_uuid}')

            for subdir_name, subdir_uuid in sorted(self._subdirs.items()):
                if '/' in subdir_name:
                    parent, _ = subdir_name.rsplit('/', maxsplit=1)
                    parent_uuid = self._subdirs[parent]
                    contents.append(f'		{subdir_uuid} = {parent_uuid}')
            contents.append('	EndGlobalSection')

        contents.append('	GlobalSection(ExtensibilityGlobals) = postSolution')
        contents.append(f'		SolutionGuid = {self.uuid}')
        contents.append('	EndGlobalSection')

        contents.append('EndGlobal')

        self.path.write_text('\n'.join(contents), encoding='utf-8')



from functools import lru_cache
import json
from pathlib import Path
from typing import Any, Optional, Iterable


def lmfind(d: list[dict[str, Any]], v: Any, key: str = 'name', default: Any = None) -> Any:
    for item in d:
        if item.get(key, ...) == v:
            return item
    return default


def lmget(d: list[dict[str, Any]], v: Any, key: str = 'name', value: str = 'value', default: Any = None) -> Any:
    item = lmfind(d, v, key, ...)
    if item is ...:
        return default
    return item[value]


class MesonVersionError(RuntimeError):
    pass


class UnconfiguredProject(RuntimeError):
    pass


class UnsuportedLanguageError(RuntimeError):
    pass


class Introspector:

    INFO_DIR = 'meson-info'
    INFO_FILE = 'meson-info.json'
    LANGUAGES = {'cpp', 'c'}
    HEADER_SUFFIXES = ('.h', '.hh', '.hpp', 'h++', '.H')

    def __init__(self, build_dir: Path):
        self.build_dir = build_dir
        self.intro_data = {}

        info_file = self.build_dir / self.INFO_DIR / self.INFO_FILE
        if not info_file.exists():
            raise UnconfiguredProject(f"{self.build_dir} does not contain a configured meson project")

        with info_file.open('r', encoding='utf8') as f:
            info_data = json.load(f)

        meson_version = info_data['meson_version']['full']
        meson_version = tuple(map(int, meson_version.split('.')))
        if meson_version < (1, 1, 99):
            raise MesonVersionError("vsgen requires at lest meson 1.2.0")

        for name, data in info_data['introspection']['information'].items():
            with (self.build_dir / self.INFO_DIR / data['file']).open('r', encoding='utf8') as f:
                self.intro_data[name] = json.load(f)

    def cpu_family(self) -> str:
        return self.intro_data["machines"]["host"]["cpu_family"]

    def is_debug(self) -> bool:
        return lmget(self.intro_data['buildoptions'], 'debug')

    def toolset(self) -> str:

        for lang in self.LANGUAGES:
            compiler = self.intro_data['compilers']['host'].get(lang)
            if compiler:
                compiler_version = compiler['version']
                version = compiler_version.split('.')
                if version[0] == '19':
                    return f'v14{version[1][0]}'
                else:
                    return f'v{int(version[0])-6}0'

        raise UnsuportedLanguageError(
            "vsgen is only compatible with projects using {} language".format(" and ".join(self.LANGUAGES))
        )

    def get_target_filename(self, target: str) -> Optional[str]:
        if target.startswith('dep:'):
            return None

        data = self._target_data(target)
        if not data:
            return None

        return data['filename'][0]

    def get_target_params(self, target: str) -> tuple[list[Path], list[str], list[str]]:
        includes = []
        macros = []
        options = []

        data = self._target_data(target)
        if not data:
            return [], [], []

        if target.startswith('dep:'):
            includes.extend(map(Path, data['include_directories']))
            for param in data['compile_args']:
                if param.startswith(('-I', '/I')):
                    includes.append(Path(param[2:]))
                elif param.startswith(('-D', '/D')):
                    macros.append(param[2:])
                else:
                    options.append(param)

        else:
            for sources in data["target_sources"]:
                lang = sources.get("language")
                if lang in self.LANGUAGES:
                    for param in sources['parameters']:
                        if param.startswith(('-I', '/I')):
                            includes.append(Path(param[2:]))
                        elif param.startswith(('-D', '/D')):
                            macros.append(param[2:])
                        else:
                            options.append(param)

        return includes, macros, options
    
    def _filter_sources(self, sources: Iterable[Path], rel_dir: Path):
        for src in sources:
            if src.is_relative_to(rel_dir):
                yield src

    def get_target_sources(self, target: str, only_relative_files: bool = False) -> list[Path]:
        data = self._target_data(target)
        if not data:
            return []

        sources = []

        if target.startswith('dep:'):
            sources.extend(map(Path, data['sources']))

        else:
            for source_item in data["target_sources"]:
                if source_item.get("linker"):
                    continue

                source_files = map(Path, source_item['sources'])
                if only_relative_files:
                    source_dir = Path(data['defined_in']).parent
                    source_files = self._filter_sources(source_files, source_dir)
                sources.extend(source_files)

                unity_sources = map(Path, source_item.get('unity_sources', []))
                if only_relative_files:
                    unity_sources = self._filter_sources(unity_sources, source_dir)
                else:
                    unity_sources = (us for us in unity_sources if not us.is_relative_to(self.build_dir))
                sources.extend(unity_sources)

        return sources

    def get_target_headers(self, target: str, only_relative_files: bool = False) -> list[Path]:
        data = self._target_data(target)
        if not data:
            return []

        if not target.startswith('dep:'):
            source_dir = Path(data['defined_in']).parent

        headers = []
        for s in data['extra_files']:
            if s.endswith(self.HEADER_SUFFIXES):
                s = Path(s)
                if target.startswith('dep:') or not only_relative_files or s.is_relative_to(source_dir):
                    headers.append(s)

        return headers

    def get_target_extra_files(self, target: str, sources_root: Path, only_relative_files: bool = False) -> list[Path]:
        data = self._target_data(target)
        if not data:
            return []
        
        source_dir = Path(data['defined_in']).parent if 'defined_in' in data else None

        extras = set()
        for s in data['extra_files']:
            if not s.endswith(self.HEADER_SUFFIXES):
                s = Path(s)
                if target.startswith('dep:') or not only_relative_files or s.is_relative_to(source_dir):
                    extras.add(s)

        if source_dir:
            # meson.build files
            for m in self.intro_data['buildsystem_files']:
                if (e := Path(m)).is_relative_to(source_dir):
                    extras.add(e)

        if not target.startswith('dep:'):
            if def_file := data.get('vs_module_defs'):
                def_file = (sources_root / def_file).resolve()
                extras.add(def_file)

        return list(sorted(extras))

    def get_target_extra_paths(self, target: str) -> list[Path]:
        if target.startswith('dep:'):
            return []

        data = self._target_data(target)
        if not data:
            return []

        target_id = data["id"]

        test_data = lmfind(self.intro_data["tests"], [target_id], key="depends")
        if not test_data:
            return []

        return list(map(Path, test_data.get("extra_paths", [])))

    def get_build_files(self) -> list[Path]:
        return list(map(Path, self.intro_data['buildsystem_files']))

    @lru_cache(maxsize=None)
    def _target_data(self, target: str) -> Optional[dict]:
        if target.startswith('dep:'):
            return lmfind(self.intro_data['dependencies'], [target[4:]], key='meson_variables', default=None)

        target_data = lmfind(self.intro_data["targets"], target)

        if target_data and target_data['type'] == "alias":

            depends = target_data.get("depends")
            if depends and len(depends) > 0:
                target_data = lmfind(self.intro_data["targets"], depends[0], key="id")

        return target_data

    def all_targets(self) -> list[str]:
        targets = {}

        for target_data in self.intro_data['targets']:
            if target_data['type'] == 'alias':
                prefix = target_data['id'].removesuffix('@run')
                for tname, tid in reversed(targets.items()):
                    if tid.startswith(prefix):
                        del targets[tname]
                        break

            targets[target_data['name']] = target_data['id']

        for dep_data in self.intro_data['dependencies']:
            if dep_data['type'] == 'internal' and dep_data['meson_variables']:
                target_name = 'dep:' + dep_data['meson_variables'][0]
                targets[target_name] = dep_data['name']

        return list(targets)



"""
GoogleTestAdapter is now part of VisualStudio, and identified as TAfGT:
https://github.com/microsoft/TestAdapterForGoogleTest
It is installed from VisualStudio Installer.
However, it may be outdated and lack features from original project.

Original GoogleTestAdapter is still maintained, and identifier as GTA:
https://github.com/csoltenborn/GoogleTestAdapter
It is installed from VisualStudio Marketplace.
However, it is not installable (yet) on VS2022:
(https://github.com/csoltenborn/GoogleTestAdapter/issues/333)

"""


from pathlib import Path
from typing import Union

import lxml.etree as ET


class RunSettingsFile:

    KNOWN_SETTINGS = {
        "AdditionalTestExecutionParam",
        "BatchForTestSetup",
        "BatchForTestTeardown",
        "BreakOnFailure",
        "CatchExceptions",
        "DebugMode",
        "TimestampMode",
        "SeverityMode",
        "SummaryMode",
        "PrefixOutputWithGta",
        "MaxNrOfThreads",
        "NrOfTestRepetitions",
        "ParallelTestExecution",
        "PrintTestOutput",
        "RunDisabledTests",
        "ShuffleTests",
        "ShuffleTestsSeed",
        "TestDiscoveryRegex",
        "AdditionalPdbs",
        "TestDiscoveryTimeoutInSeconds",
        "WorkingDir",
        "PathExtension",
        "EnvironmentVariables",
        "TraitsRegexesAfter",
        "TraitsRegexesBefore",
        "TestNameSeparator",
        "ParseSymbolInformation",
        "UseNewTestExecutionFramework",
        "KillProcessesOnCancel",
        "ExitCodeTestCase",
        "MissingTestsReportMode",
    }
    TRAITS_REGEX_SEPARATOR = "//||//"

    def __init__(self, path: Path):
        self.path = path

        self.solution_settings = {}
        self.project_settings = {}

    def add_solution_setting(self, name: str, value: Union[str, bool, list[str]]):
        if name not in self.KNOWN_SETTINGS:
            raise ValueError(f'Invalid setting {name}')
        self.solution_settings[name] = value

    def add_project_setting(self, name: str, value: Union[str, bool, list[str]], regex: str):
        if name not in self.KNOWN_SETTINGS:
            raise ValueError(f'Invalid setting {name}')
        s = self.project_settings.setdefault(regex, {})
        s[name] = value

    def _write_setting(self, parent: ET.Element, items: dict[str, Union[str, bool, list[str]]]):
        for key, value in items.items():
            if isinstance(value, bool):
                value = str(value).lower()

            elif isinstance(value, list):
                value = self.TRAITS_REGEX_SEPARATOR.join(value)

            ET.SubElement(parent, key).text = value

    def write(self):
        root = ET.Element("RunSettings")
        gta = ET.SubElement(root, "GoogleTestAdapterSettings")

        if self.solution_settings:
            sol = ET.SubElement(gta, "SolutionSettings")
            solset = ET.SubElement(sol, "Settings")
            self._write_setting(solset, self.solution_settings)

        if self.project_settings:
            prj = ET.SubElement(gta, "ProjectSettings")
            for regex, settings in self.project_settings.items():
                prjset = ET.SubElement(prj, "Settings", ProjectRegex=regex)
                self._write_setting(prjset, settings)

        self.path.write_bytes(ET.tostring(root, encoding='UTF-8', xml_declaration=True, pretty_print=True))



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path


@dataclass(eq=True, unsafe_hash=True)
class BaseConfig:
    name: str
    arch: str

    def __str__(self) -> str:
        return f'{self.name}|{self.arch}'

    @classmethod
    def from_str(cls, config: str) -> BaseConfig:
        return cls(*config.split('|'))

    @property
    def condition(self) -> str:
        return f"'$(Configuration)|$(Platform)'=='{self!s}'"

    def __lt__(self, other: Configuration) -> bool:
        return (self.name, self.arch) < (other.name, other.arch)


@dataclass(eq=True)
class Configuration(BaseConfig):

    is_debug: bool
    toolset: str
    builddir: Path
    use_env_include: bool = False


@dataclass
class BuildParams:

    output: str | None = None
    include_directories: list[Path] = field(default_factory=list)
    macros: list[str] = field(default_factory=list)
    additional_options: list[str] = field(default_factory=list)
    search_paths: list[Path] = field(default_factory=list)



from pathlib import Path

from cleo.commands.command import Command
from cleo.helpers import argument, option

from ._pathconfig import PathConfig, pathconfig
from ..api import Generator


class SolutionCommand(Command):
    name = "solution"
    description = "Generate one Visual Studio solution (.sln file)."
    arguments = [
        argument("name", description="solution name or path of generated solution"),
        argument(
            "project", description="projects to include ([target:][path/to/]project)", multiple=True, optional=True
        ),
    ]
    options = [
        option("projectdir", "p", description="where to put projects (relative to outputdir)", flag=False),
        option("update", "u", description="update existing solution and projects", flag=True),
    ] + PathConfig.options

    def handle(self) -> int:
        try:
            paths = pathconfig(self)
        except SystemExit as e:
            if e.code:
                return e.code

        projectdir = self.option("projectdir")
        if projectdir:
            projectdir = Path(projectdir)
            if projectdir.is_absolute():
                self.line_error("projectdir must be a relative path", "error")
                return 2
        else:
            projectdir = ""

        projects = self.argument("project")
        if projects:
            for p in projects:
                if ":" in p:
                    target, p = p.rsplit(p, maxsplit=1)
                else:
                    target = None
                pp = Path(p)
                if pp.name == p:
                    subdir = True
                else:
                    subdir = pp.parent
                projects.append((pp.stem, subdir, target))
        else:
            projects = None

        self.info("Generating VisualStudio solution...")
        gen = Generator(paths.output_path, paths.source_path, projectdir)

        update = self.option("update")
        for build_path, config_name in paths.build_paths.items():
            gen.set_config(build_path, config_name)
            sln = gen.solution(paths.name, update)

            if projects:
                for name, subdir, target in projects:
                    prj = gen.project(name, target, subdir, update)
                    prj.write()
                    sln.add_project(prj)
            else:
                for target in gen.all_targets():
                    prj = gen.project(target, subdir=False, update=update)
                    prj.write()
                    sln.add_project(prj)

            all_prj = gen.project("ALL", "all", False, update)
            all_prj.write()
            sln.add_project(all_prj, build_solution_target=True)

            sln.write()
            update = True

        return 0



from __future__ import annotations

import os.path
from pathlib import Path
import re
from typing import Optional, Union

import yaml

from .api import Generator
from .introspector import Introspector, UnconfiguredProject
from .runsettings import RunSettingsFile
from .sln import SolutionFile
from .vcxproj import VcxProj


class ConfigFile:

    SPECIAL_TARGETS = {'all', 'test'}

    def __init__(self, path: Path):
        self.path = path
        self.data = {'projects': {'*': False}}

    def load(self) -> ConfigFile:
        with self.path.open('r', encoding='utf8') as cf:
            self.data.update(yaml.safe_load(cf) or {})
        return self

    def analyse(self, basepath: Path) -> dict:
        results = {}

        if not self.data.get('configs'):
            return {'error': 'no configs defined'}

        targets = set()
        for config_path in self.data['configs']:
            try:
                introspector = Introspector(basepath / config_path)
                targets.update(introspector.all_targets())
            except UnconfiguredProject as e:
                print(e)

        if not targets:
            return {'error': 'no targets found. Is project configured?'}

        if self.data['projects'].get('*'):
            results.update(self._analyse_targets(targets))

        return results

    def generate(self, basepath: Path, outputpath: Optional[Path] = None, update: bool = True):

        use_include_from_env = self.data.get('include_from_env', False)
        if use_include_from_env == 'auto':
            use_include_from_env = os.environ.get('INCLUDE', None) is not None

        if not outputpath:
            outputpath = Path(self.data.get('outputdir', '.'))

        gen = Generator(outputpath, basepath, self.data.get('projectdir', ''))
        gen.cpu_arch = self.data.get('archs', {})

        all_projects = {}
        solutions = {}
        for config_path, config_name in self.data.get('configs', {}).items():
            if not gen.set_config(config_path, config_name, use_include_from_env):
                continue

            projects = self._generate_projects(gen, update)
            for p in projects:
                all_projects[p.uuid] = p
            solutions = self._generate_solutions(gen, projects, update)
            update = True

        self._generate_runsettings(gen, solutions, all_projects.values())

    def _analyse_targets(self, targets: set[str]) -> dict:
        listed_targets = set()
        for project_name, target in self.data['projects'].items():
            if project_name == '*':
                continue
            if isinstance(target, bool):
                listed_targets.add(project_name)
            elif isinstance(target, str):
                listed_targets.add(target)
            else:
                target = target.get('target', project_name)
                if isinstance(target, bool):
                    target = project_name

                listed_targets.add(target)

        unlisted_targets = targets - listed_targets
        missing_targets = listed_targets - targets - self.SPECIAL_TARGETS

        result = {}
        if unlisted_targets:
            result['unlisted_targets'] = list(unlisted_targets)
        if missing_targets:
            result['unexisting_targets'] = list(missing_targets)
        return result

    def _analyse_solutions_with_unexisting_projects(self):
        pass

    def _analyse_solutions_without_build(self):
        pass

    def _generate_projects(self, gen: Generator, update: bool) -> list[VcxProj]:
        ci_targets = self.data.get('case_insensitive_targets', False)
        orf = self.data.get('only_relative_files', False)

        projects = []
        all_targets = set(gen.all_targets())

        default_config = {
            'target': True,
            'subdir': '',
        }
        all_config = self.data.get('projects', {}).get('*')
        if isinstance(all_config, dict):
            default_config.update(all_config)

        for name, config in self.data.get('projects', {}).items():
            if name == '*':
                continue

            if config is False:
                if name in all_targets:
                    all_targets.remove(name)
                elif ci_targets and name.lower() in all_targets:
                    all_targets.remove(name.lower())
                elif 'dep:' + name in all_targets:
                    all_targets.remove('dep:' + name)
                elif ci_targets and 'dep:' + name.lower() in all_targets:
                    all_targets.remove('dep:' + name.lower())
                continue

            if config is True:
                config = default_config
            elif isinstance(config, str):
                config = default_config | {'target': config}
            else:
                config = default_config | config

            target = config['target']
            if target is True:
                target = name

                if target not in all_targets:
                    if ci_targets and target.lower() in all_targets:
                        target = target.lower()
                    elif 'dep:' + target in all_targets:
                        target = 'dep:' + target
                    elif ci_targets and 'dep:' + target.lower() in all_targets:
                        target = 'dep:' + target.lower()

            elif target is False:
                continue

            if isinstance(target, list):
                for t in target.copy():
                    if t not in all_targets and t not in self.SPECIAL_TARGETS:
                        target.remove(t)
                if not target:
                    continue

                for t in target:
                    if t in all_targets:
                        all_targets.remove(t)

                target = " ".join(target)

            else:
                if target in all_targets:
                    all_targets.remove(target)
                elif target not in self.SPECIAL_TARGETS:
                    target = False  # dummy target

            project = gen.project(name, target, config['subdir'], update, orf).write()
            projects.append(project)

        if self.data.get('projects', {}).get('*'):
            for target in all_targets:
                if target.startswith('dep:'):
                    continue
                project = gen.project(target, target, default_config['subdir'], update).write()
                projects.append(project)

        return projects

    def _generate_solutions(self, gen: Generator, projects: list[VcxProj], update: bool) -> dict[str, SolutionFile]:
        solutions = {}
        for name, slnconfig in self.data.get('solutions', {}).items():
            sln = gen.solution(name, update)

            build_solution = slnconfig.get('build_solution', [])
            if isinstance(build_solution, str):
                build_solution = [build_solution]

            structure = parse_project_struture(slnconfig.get('projects', {}))
            for project in projects:
                subdir = structure.get(project.name, structure.get('*'))
                if subdir is not None:
                    sln.add_project(project, subdir, project.name in build_solution)

            sln.write()
            solutions[name] = sln
        return solutions

    def _generate_runsettings(self, gen: Generator, solutions: dict[str, SolutionFile], projects: list[VcxProj]):
        for name, slnconfig in self.data.get('solutions', {}).items():
            if runsettings := slnconfig.get('runsettings'):
                sln = solutions[name]
                rsfile = RunSettingsFile(sln.path.with_suffix('.gta.runsettings'))
                prdict = {project.uuid: project for project in projects}

                if runsettings is True:
                    runsettings = [{".*\\.exe": True}]

                all_keys = {"*", ".*"}
                for rsconfig in runsettings:
                    for test_regex, rsoptions in rsconfig.items():
                        if test_regex in all_keys:
                            if rsoptions is True:
                                rsoptions = {}
                            for key, value in rsoptions.items():
                                rsfile.add_solution_setting(key, value)
                    for k in all_keys:
                        rsconfig.pop(k, None)

                for project_id in sln.projects:
                    project = prdict[project_id]

                    for config_name, config in project.configs.items():
                        params = project.params[config_name]

                        if not params.output:
                            continue

                        output = os.path.relpath(
                            (project.path / config.builddir / params.output).resolve(), gen.outputdir
                        )

                        is_matching = False
                        for rsconfig in runsettings:
                            for test_regex, rsoptions in rsconfig.items():
                                if re.fullmatch(test_regex, output):
                                    is_matching = True
                                    break
                            if is_matching:
                                break
                        if not is_matching:
                            continue

                        if rsoptions is True:
                            rsoptions = {}

                        output = output.removeprefix('..')
                        while output.startswith('\\..'):
                            output = output.removeprefix('\\..')

                        regex = ".*" + output.replace('\\', '\\\\')

                        for key, value in rsoptions.items():
                            rsfile.add_project_setting(key, value, regex)

                        pathext = ';'.join(
                            str((project.path / config.builddir / p).resolve()) for p in params.search_paths
                        )
                        if pathext:
                            rsfile.add_project_setting("PathExtension", pathext, regex)

                rsfile.write()


def parse_project_struture(structure: Union[dict, list], current='') -> dict[str, str]:
    result = {}

    if isinstance(structure, list):
        for item in structure:
            if isinstance(item, str):
                result[item] = current
            else:
                result.update(parse_project_struture(item, current))
    else:
        for subdir, item in structure.items():
            if item is False:
                result[subdir] = None
            elif item is True:
                result[subdir] = current
            elif isinstance(item, str):
                result[item] = f'{current}/{subdir}'
            else:
                result.update(parse_project_struture(item, f'{current}/{subdir}'))

    return result



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from .configuration import Configuration, BuildParams
from .introspector import Introspector
from .vcxproj import VcxProj
from .sln import SolutionFile


if TYPE_CHECKING:
    from os import PathLike

    PathOrStr = Union[PathLike, str]


class Generator:
    def __init__(self, vspath: PathOrStr, basedir: PathOrStr = ".", projectdir: PathOrStr = "projects") -> None:
        self.basedir: Path = Path(basedir).absolute()
        self.outputdir: Path = self.basedir / vspath
        self.projectdir: Path = self.outputdir / projectdir

        # cpu_family to arch conversion
        # you may override or update this dict
        self.cpu_arch: dict[str, str] = {
            'x86_64': 'x64',
            'x86': 'Win32',
        }

        self._intro: Optional[Introspector] = None
        self._config: Optional[Configuration] = None

    def set_config(self, builddir: PathOrStr, name: Optional[str] = None, use_include_from_env: bool = False) -> bool:
        try:
            self._intro = Introspector(self.basedir / builddir)
        except RuntimeError:
            return False

        cpu_family = self._intro.cpu_family()
        arch = self.cpu_arch.get(cpu_family, cpu_family)

        self._config = Configuration(
            name or self._intro.build_dir.name,
            arch,
            self._intro.is_debug(),
            self._intro.toolset(),
            self._intro.build_dir,
            use_include_from_env,
        )
        return True

    @property
    def introspector(self) -> Introspector:
        if self._intro is None:
            raise RuntimeError("Generator not configurated. Call `set_config()` first.")
        return self._intro

    @property
    def builddir(self) -> Path:
        return self.introspector.build_dir

    @property
    def config(self) -> Configuration:
        if self._config is None:
            raise RuntimeError("Generator not configurated. Call `set_config()` first.")
        return self._config

    def all_targets(self) -> list[str]:
        return self.introspector.all_targets()

    def project(
        self,
        name: str,
        target: Union[str, bool, None] = True,
        subdir: Union[PathOrStr, bool] = False,
        update: bool = True,
        only_relative_files: bool = False
    ) -> VcxProj:
        project_path = self.projectdir
        if subdir is True:
            project_path /= name
        elif subdir:
            project_path /= subdir

        vcxproj = VcxProj(name, project_path)
        if update:
            vcxproj.load(required=False)
        else:
            vcxproj.remove_user_file()

        vcxproj.add_config(self.config)

        if target:
            if target is True:
                target = name

            output = self.introspector.get_target_filename(target)
            extra_paths = self.introspector.get_target_extra_paths(target)

        if not target:
            build_params = BuildParams()
        elif target.startswith('dep:'):
            build_params = BuildParams(
                None,
                *self.introspector.get_target_params(target)
            )
        elif output:
            output = Path(output).relative_to(self.builddir)

            build_params = BuildParams(
                str(output),
                *self.introspector.get_target_params(target),
                extra_paths,
            )
        else:
            build_params = BuildParams(output=target)

        vcxproj.add_build_params(str(self.config), build_params)

        if target:
            if output or target.startswith('dep:'):
                vcxproj.add_sources(self.introspector.get_target_sources(target, only_relative_files))
                vcxproj.add_headers(self.introspector.get_target_headers(target, only_relative_files))
                vcxproj.add_extra_files(self.introspector.get_target_extra_files(target, self.basedir, only_relative_files))
            elif target == 'all':
                vcxproj.add_extra_files(self.introspector.get_build_files())

        return vcxproj

    def solution(self, name: str, update: bool = True) -> SolutionFile:
        sln = SolutionFile(self.outputdir / name)

        if update:
            sln.load_configs()

        return sln



from cleo.commands.command import Command
from cleo.helpers import argument, option

from ._pathconfig import PathConfig, pathconfig
from ..api import Generator


class ProjectCommand(Command):
    name = "project"
    description = "Generate one Visual Studio project (.vcxproj file)."
    arguments = [argument("name", description="project name or path of generated project")]
    options = [
        option("target", "t", description="build target to use (default is same as project name)", flag=False),
        option("update", "u", description="update existing project", flag=True),
    ] + PathConfig.options

    def handle(self) -> int:
        try:
            paths = pathconfig(self)
        except SystemExit as e:
            if e.code:
                return e.code

        self.info("Generating VisualStudio project...")
        gen = Generator(paths.output_path, paths.source_path, "")

        update = self.option("update")
        for build_path, config_name in paths.build_paths.items():
            gen.set_config(build_path, config_name)
            prj = gen.project(paths.name, self.option("target"), paths.project_dir, update)
            prj.write()
            update = True

        return 0



from pathlib import Path
from typing import Optional

from cleo.commands.command import Command
from cleo.helpers import option


class PathConfig:

    options = [
        option("sourcedir", "s", description="path to source dir (default is current dir)", flag=False),
        option("outputdir", "o", description="where to write projects and solutions", flag=False),
        option(
            "builddir",
            "b",
            description="paths to build dir ([configname:]path/to/builddir) (relative to sourcedir)",
            flag=False,
            multiple=True,
        ),
    ]

    def __init__(self, name: str, sourcedir: Optional[str], builddir: Optional[list[str]], outputdir: Optional[str]):
        self.source_path = Path(sourcedir or ".").resolve()

        namepath = Path(name)
        self.name = namepath.stem

        self.build_paths = {}
        builddirs = builddir or []

        for b in builddirs:
            if ':' in b[2:]:  # do not split if b is "C:\..."
                config_name, build_path = b.split(":", maxsplit=1)

            else:
                config_name = None
                build_path = b

            build_path = self.source_path / build_path
            self.build_paths[build_path] = config_name

        if outputdir:
            self.output_path = Path(outputdir).resolve().absolute()
            self.project_dir = (self.output_path / namepath.parent).resolve()
            if self.project_dir.is_relative_to(self.output_path):
                self.project_dir = self.project_dir.relative_to(self.output_path)
            if self.project_dir == Path("."):
                self.project_dir = False

        else:
            self.output_path = namepath.parent.resolve().absolute()
            self.project_dir = False


def pathconfig(cmd: Command) -> PathConfig:
    pc = PathConfig(cmd.argument("name"), cmd.option("sourcedir"), cmd.option("builddir"), cmd.option("outputdir"))

    if not pc.source_path.is_dir():
        cmd.line_error("sourcedir must be a directory", "error")
        raise SystemExit(1)

    if not pc.build_paths:
        cmd.line_error("You must provide at least one builddir", "error")
        raise SystemExit(2)

    if pc.project_dir and pc.project_dir.is_absolute():
        cmd.line_error("name should be relative to output path", "error")
        raise SystemExit(1)

    return pc



import pkg_resources

from cleo.application import Application

from .generate import GenerateCommand
from .project import ProjectCommand
from .solution import SolutionCommand

version = pkg_resources.get_distribution('meson-vs-gen').version
application = Application("vsgen", version)
application.add(GenerateCommand())
application.add(ProjectCommand())
application.add(SolutionCommand())


def main() -> int:
    return application.run()



from pathlib import Path

from cleo.commands.command import Command
from cleo.helpers import argument, option

from ..configfile import ConfigFile
from ..introspector import MesonVersionError


class GenerateCommand(Command):
    name = "generate"
    description = "Generate solutions and projects from config file."
    arguments = [argument("configfile", description="yaml config file")]
    options = [
        option("sourcedir", "s", description="path to source dir (default is current dir)", flag=False),
        option("outputdir", "o", description="where to write projects and solutions", flag=False),
        option("update", "u", description="update existing solution and projects", flag=True),
    ]

    def handle(self) -> int:
        configfile = Path(self.argument("configfile"))
        if not configfile.is_file():
            self.line_error(f"{configfile} does not exist", "error")
            return 1

        source_path = Path(self.option("sourcedir") or ".").resolve()
        if not source_path.is_dir():
            self.line_error("sourcedir must be a directory", "error")
            return 1

        output_path = self.option("outputdir")
        if output_path:
            output_path = Path(output_path)

        self.info("Generating VisualStudio solution...")
        cf = ConfigFile(configfile).load()
        try:
            cf.analyse(source_path)
        except MesonVersionError as e:
            self.error(str(e))
            return 1
        cf.generate(source_path, output_path, update=self.option("update"))
        return 0



from __future__ import annotations

from collections.abc import Iterable
import os.path
from pathlib import Path
import re
from typing import Optional
import uuid

import lxml.etree as ET

from .configuration import Configuration, BuildParams


class VcxProj:

    NS: str = 'http://schemas.microsoft.com/developer/msbuild/2003'
    EXT = '.vcxproj'
    FILTER_EXT = EXT + '.filters'
    USER_EXT = EXT + '.user'

    def __init__(self, name: str, path: Path):
        self.name: str = name
        self.path: Path = path
        self.uuid: str = f'{{{str(uuid.uuid4())}}}'.upper()

        self.configs: dict[str, Configuration] = {}
        self.params: dict[str, BuildParams] = {}

        self._files: dict[Path, str] = {}

    @property
    def filepath(self) -> Path:
        return self.path / (self.name + self.EXT)

    @property
    def filterspath(self) -> Path:
        return self.path / (self.name + self.FILTER_EXT)

    @property
    def userpath(self) -> Path:
        return self.path / (self.name + self.USER_EXT)

    def _load_configs(self, project_xml: ET.Element):
        self.uuid = project_xml.find(f'.//{{{self.NS}}}ProjectGuid').text

        for config_node in project_xml.iterfind(f'.//{{{self.NS}}}ProjectConfiguration'):
            config = config_node.findtext(f"{{{self.NS}}}Configuration")
            arch = config_node.findtext(f"{{{self.NS}}}Platform")

            bc = Configuration(config, arch, False, "", Path())
            self.configs[str(bc)] = bc
            self.params[str(bc)] = BuildParams()

        for config_node in project_xml.iterfind(f'.//{{{self.NS}}}PropertyGroup'):
            condition = config_node.get("Condition")
            config = None
            for bc in self.configs.values():
                if bc.condition == condition:
                    config = bc
                    params = self.params[str(bc)]
                    break

            if config is None:
                continue

            label = config_node.get('Label')

            if label == 'UserMacros':
                config.builddir = Path(config_node.find(f'{{{self.NS}}}BuildDir').text)
                if search_paths := self._read_opt_node(config_node, 'DllPaths'):
                    params.search_paths = list(map(Path, search_paths.split(';')))

            elif label == "Configuration":
                config.toolset = self._read_opt_node(config_node, 'PlatformToolset')
                config.is_debug = config_node.find(f'{{{self.NS}}}UseDebugLibraries').text == 'true'

            elif label is None:
                if include_directories := self._read_opt_node(config_node, 'NMakeIncludeSearchPath'):
                    params.include_directories = list(map(Path, include_directories.split(';')))
                if macros := self._read_opt_node(config_node, 'NMakePreprocessorDefinitions'):
                    params.macros = macros.split(' ')
                if additional_options := self._read_opt_node(config_node, 'AdditionalOptions'):
                    params.additional_options = additional_options.split(';')
                if self._read_opt_node(config_node, 'IncludePath') == '$(INCLUDE)':
                    config.use_env_include = True

                build_cmd = self._read_opt_node(config_node, 'NMakeBuildCommandLine', '')
                m = re.fullmatch(r'ninja -C (\S+) (\S+)', build_cmd)
                params.output = m[2] if m else None

    def _load_files(self, project_xml: ET.Element):
        for filetype in ("ClCompile", "ClInclude", "None"):
            for file_node in project_xml.iterfind(f'.//{{{self.NS}}}{filetype}'):
                self._files[Path(self.filepath.parent, file_node.get("Include")).resolve()] = filetype

    def load(self, required=True) -> None:
        if not self.filepath.exists():
            if required:
                raise FileNotFoundError(f'{self.filepath} does not exist')
            else:
                return

        project_xml = ET.parse(self.filepath).getroot()
        self._load_configs(project_xml)
        self._load_files(project_xml)

    def _read_opt_node(self, parent: ET.Element, name: str, default: Optional[str] = None) -> str | None:
        node = parent.find(f'{{{self.NS}}}{name}')
        if node is None:
            return default
        return node.text or default

    def remove_user_file(self) -> None:
        self.userpath.unlink(missing_ok=True)

    def add_config(self, config: Configuration):
        self.configs[str(config)] = config

    def add_build_params(self, config: str, params: BuildParams):
        self.params[config] = params

    def write(self) -> VcxProj:
        self.path.mkdir(parents=True, exist_ok=True)
        self._write_project_file()
        self._write_filters_file()
        if not self.userpath.exists():
            self._write_user_file()
        return self  # for chaining

    def _xml_project(self, **kwargs: str) -> ET.Element:
        nsmap = {None: self.NS}
        return ET.Element('Project', nsmap=nsmap, attrib=kwargs)

    @staticmethod
    def _write_xml(path: Path, xml_root: ET.Element) -> None:
        path.write_bytes(ET.tostring(xml_root, encoding='UTF-8', xml_declaration=True, pretty_print=True))

    def _write_project_file(self) -> None:
        root = self._xml_project(DefaultTargets='Build')

        project_configurations = ET.SubElement(root, 'ItemGroup', attrib={'Label': 'ProjectConfigurations'})
        for config in sorted(self.configs.values()):
            pc = ET.SubElement(project_configurations, 'ProjectConfiguration', attrib={'Include': str(config)})
            ET.SubElement(pc, 'Configuration').text = config.name
            ET.SubElement(pc, 'Platform').text = config.arch

        globals_group = ET.SubElement(root, 'PropertyGroup', attrib={'Label': 'Globals'})
        ET.SubElement(globals_group, 'VCProjectVersion').text = '15.0'
        ET.SubElement(globals_group, 'ProjectGuid').text = self.uuid
        ET.SubElement(globals_group, 'Keyword').text = 'MakeFileProj'

        ET.SubElement(root, 'Import', attrib={'Project': '$(VCTargetsPath)\\Microsoft.Cpp.Default.props'})

        for config in sorted(self.configs.values()):
            conf = ET.SubElement(
                root, 'PropertyGroup', attrib={'Label': 'Configuration', 'Condition': config.condition}
            )
            ET.SubElement(conf, 'ConfigurationType').text = 'Makefile'
            ET.SubElement(conf, 'UseDebugLibraries').text = str(config.is_debug).lower()
            ET.SubElement(conf, 'PlatformToolset').text = config.toolset

        ET.SubElement(root, 'Import', attrib={'Project': '$(VCTargetsPath)\\Microsoft.Cpp.props'})
        ET.SubElement(root, 'ImportGroup', attrib={'Label': 'ExtensionSettings'})
        ET.SubElement(root, 'ImportGroup', attrib={'Label': 'Shared'})

        for config in sorted(self.configs.values()):
            prop_sheets = ET.SubElement(
                root, 'ImportGroup', attrib={'Label': 'PropertySheets', 'Condition': config.condition}
            )
            ET.SubElement(
                prop_sheets,
                'Import',
                attrib={
                    'Label': 'LocalAppDataPlatform',
                    'Project': '$(UserRootDir)\\Microsoft.Cpp.$(Platform).user.props',
                    'Condition': "exists('$(UserRootDir)\\Microsoft.Cpp.$(Platform).user.props')",
                },
            )

        for config in sorted(self.configs.values()):
            params = self.params[str(config)]

            user_macros = ET.SubElement(
                root, 'PropertyGroup', attrib={'Label': 'UserMacros', 'Condition': config.condition}
            )
            ET.SubElement(user_macros, 'BuildDir').text = self._relpath(config.builddir)
            ET.SubElement(user_macros, 'DllPaths').text = ';'.join(
                self._relpath(p, config.builddir) for p in params.search_paths
            )

        for config in sorted(self.configs.values()):
            params = self.params[str(config)]

            compile_cmd = ET.SubElement(root, 'PropertyGroup', attrib={'Condition': config.condition})
            if params.output is not None:
                ET.SubElement(compile_cmd, 'NMakeBuildCommandLine').text = f'ninja -C "$(BuildDir)" {params.output}'
                if params.output:
                    ET.SubElement(compile_cmd, 'NMakeOutput').text = self._relpath(config.builddir / params.output)
                ET.SubElement(
                    compile_cmd, 'NMakeReBuildCommandLine'
                ).text = f'ninja -C "$(BuildDir)" clean && ninja -C "$(BuildDir)" {params.output}'
            ET.SubElement(compile_cmd, 'NMakeCleanCommandLine').text = 'ninja -C "$(BuildDir)" clean'
            if params.macros:
                ET.SubElement(compile_cmd, 'NMakePreprocessorDefinitions').text = ';'.join(params.macros)
            if params.include_directories:
                ET.SubElement(compile_cmd, 'NMakeIncludeSearchPath').text = ';'.join(
                    map(self._relpath, params.include_directories)
                )
            if params.additional_options:
                ET.SubElement(compile_cmd, 'AdditionalOptions').text = ' '.join(params.additional_options)
            ET.SubElement(compile_cmd, 'LocalDebuggerWorkingDirectory').text = '$(BuildDir)'
            ET.SubElement(compile_cmd, 'LocalDebuggerEnvironment').text = 'PATH=$(DllPaths)'

            if config.use_env_include:
                ET.SubElement(compile_cmd, 'IncludePath').text = '$(INCLUDE)'

        cat_nodes: dict[str, ET.Element] = {}
        for filename, category in sorted(self._files.items()):
            node = cat_nodes.get(category)
            if node is None:
                node = ET.SubElement(root, 'ItemGroup')
                cat_nodes[category] = node
            ET.SubElement(node, category, attrib={'Include': self._relpath(filename)})

        ET.SubElement(root, 'Import', attrib={'Project': '$(VCTargetsPath)\\Microsoft.Cpp.targets'})
        ET.SubElement(root, 'ImportGroup', attrib={'Label': 'ExtensionTargets'})

        self._write_xml(self.filepath, root)

    def _find_source_dir(self) -> Path | None:
        if self._files:
            f = next(iter(self._files))
            for p in f.parents:
                if all(o.is_relative_to(p) for o in self._files):
                    return p
        return None

    def _write_filters_file(self) -> None:
        source_dir = self._find_source_dir()

        def _dirname(file: Path) -> str:
            return str(file.relative_to(source_dir).parent if source_dir else file.parent)

        subdirs = set()
        for file in sorted(self._files):
            dirname = _dirname(file)
            if dirname != '.':
                subdirs.add(dirname)
                for p in Path(dirname).parents:
                    if str(p) != '.' and str(p) not in subdirs:
                        subdirs.add(str(p))
                        continue
                    break

        root = self._xml_project(ToolsVersion='4.0')
        filter_list = ET.SubElement(root, "ItemGroup")
        for dirname in sorted(subdirs):
            ET.SubElement(filter_list, "Filter", attrib={"Include": dirname})

        file_list = ET.SubElement(root, "ItemGroup")
        for filename, category in sorted(self._files.items()):
            file_node = ET.SubElement(file_list, category, attrib={"Include": self._relpath(filename)})
            dirname = _dirname(filename)
            if dirname != '.':
                ET.SubElement(file_node, "Filter").text = dirname

        self._write_xml(self.filterspath, root)

    def _write_user_file(self) -> None:
        root = self._xml_project(ToolsVersion='Current')
        ET.SubElement(root, 'PropertyGroup')
        self._write_xml(self.userpath, root)

    def _relpath(self, f: Path, base: Path | None = None) -> str:
        if f.is_absolute():
            return os.path.relpath(f, base or self.path)
        else:
            return str(f)

    def add_sources(self, sources: Iterable[Path]) -> None:
        for f in sources:
            self._files[f] = 'ClCompile'

    def add_headers(self, headers: Iterable[Path]) -> None:
        for f in headers:
            self._files[f] = 'ClInclude'

    def add_extra_files(self, extras: Iterable[Path]) -> None:
        for f in extras:
            self._files[f] = 'None'
